module ConfigFile(
    loadStdConfiguration,
    configLookupBool,
    configLookup,
    configLookupList,
    -- new interface
    configSignal,
    reloadConfigFiles,
    mapConfig,
    configFile,
    configEnv,
    configDefault,
    configShow,
    parseConfigFile
    
    
    ) where

import Char
import System
import System.IO.Unsafe

import Signal
import Control.Monad.Trans
import Control.Monad.Monoid


-- | emitted when the configuration settings are changed.
configSignal :: Signal ()
configSignal = signal_ $ svarSignal config_files_var

{-# NOTINLINE config_files_var #-}
config_files_var :: SVar [(String,[(String,String)])]
config_files_var = unsafePerformIO (newSVar [])

-- | reload all configuration files emit a config signal.
reloadConfigFiles :: IO ()
reloadConfigFiles = writeSVar config_files_var []

type Config = String -> IO [(String,(String,String))]

basicLookup n cl k = return [ (n,(k,v)) | (k',v) <- cl, k == k']

configDefault :: [(String,String)] -> Config 
configDefault cl k = basicLookup "default" cl k

configFile :: String -> Config
configFile fn k = do
    cf <- readSVar config_files_var
    case lookup fn cf of 
	Just cl -> basicLookup fn cl k
	Nothing -> do 
	    cl <- catch (fmap parseConfigFile $ readFile fn) (\_ -> return []) 
	    modifySVar config_files_var ((fn,cl):)
	    basicLookup fn cl k

configEnv :: Config
configEnv k = do
    ev <- catch (fmap return $ getEnv k) (\_ -> return [])
    return $ fmap (\v -> ("enviornment", (k,v))) ev

mapConfig :: (String -> String) -> Config -> Config
mapConfig f c s = c (f s) 
    
instance Monoid Config where
    mempty _ =  return []
    mappend c1 c2 s = do
	x <- c1 s 
	y <- c2 s
	return (x ++ y)

configShow :: [String] -> Config -> IO String
configShow ss c = do
    v <- mapM c ss
    return $ unlines $ map p $ zip ss v where
	p (k,((w,(k',v))):_) = k ++ " " ++ v ++ "\n#  in " ++ w ++ 
	    if k' /= k then " as " ++ k' else ""
	p (k,[]) = "#" ++ k ++ " Not Found."
    

-- types of config sources:
-- enviornment
-- enviornment after transformation of query
-- file 
-- default


type ConfigFile = [(String, String)]


getConfiguration :: String -> IO ConfigFile
getConfiguration s = do
    e <- getEnv "HOME"
    c <- readFile (e ++ "/" ++ s)
    return $ parseConfigFile c
    


isPropName c = isAlphaNum c || c `elem` "-_"

parseConfigFile :: String -> ConfigFile
parseConfigFile s = concatMap bl (fixup $ lines (uncomment s)) where
    uncomment ('#':xs) = uncomment (dropWhile (/= '\n') xs)
    uncomment ('-':'-':xs) = uncomment (dropWhile (/= '\n') xs)
    uncomment (x:xs) = x:uncomment xs
    uncomment [] = []
    fixup (x:y@(c:_):xs) | isSpace c = fixup ((x ++ y):xs)
    fixup (x:xs) = x: fixup xs
    fixup [] = []
    bl s = let (n,r) = span isPropName (dropWhile isSpace s) in
	if null n then [] else [(n,dropWhile isSpace r)]


{-# NOINLINE config_file_var #-}
config_file_var :: SVar ConfigFile
config_file_var = unsafePerformIO $ newSVar []


loadStdConfiguration :: String -> IO ()
loadStdConfiguration f = do
    c <- getConfiguration f
    writeSVar config_file_var c
    return ()


configLookupBool k = do
    vs <- readSVar config_file_var
    case lookup k vs of 
	Just s | cond -> return True where
	    cond = s' == "true" || s' == "yes" || s' == "on" || s' == "y" || s' == "t"
	    s' = map toLower s
	_ -> return False
    

configLookup k = do
    vs <- readSVar config_file_var
    return $ lookup k vs


configLookupList k = do
    vs <- readSVar config_file_var
    return $ [ y| (x,y) <- vs, x == k]

